Imports System.Runtime.InteropServices
Imports System.IO
Imports System.ComponentModel
Module SetPerms3
    ' Type of Securable Object we are operating in this sample code
    Private Const SE_FILE_OBJECT As Integer = 1&

    ' The Security Information constants required
    Private Const DACL_SECURITY_INFORMATION As Integer = 4&
    Private Const SET_ACCESS As Integer = 2&

    ' Standard access rights extracted from WinNT.h
    Private Const SYNCHRONIZE As Integer = &H100000
    Private Const READ_CONTROL As Integer = &H20000
    Private Const WRITE_DAC As Integer = &H40000
    Private Const WRITE_OWNER As Integer = &H80000
    Private Const STANDARD_RIGHTS_READ As Integer = (READ_CONTROL)
    Private Const STANDARD_RIGHTS_WRITE As Integer = (READ_CONTROL)
    Private Const DELETE As Integer = &H10000
    Private Const DELETE_CHILD As Integer = &H40
    Private Const ALL_ACCESS As Integer = &HF0000 Or SYNCHRONIZE Or &H1FF

    ' Generic access rights extracted from WinNT.h
    Private Const GENERIC_ALL As Integer = &H10000000
    Private Const GENERIC_EXECUTE As Integer = &H20000000
    Private Const GENERIC_READ As Integer = &H80000000
    Private Const GENERIC_WRITE As Integer = &H40000000

    ' Inheritance Flags
    Private Const CONTAINER_INHERIT_ACE As Integer = &H2
    Private Const OBJECT_INHERIT_ACE As Integer = &H1

    'typedef struct _TRUSTEE {
    '  PTRUSTEE pMultipleTrustee;
    '  MULTIPLE_TRUSTEE_OPERATION MultipleTrusteeOperation;
    '  TRUSTEE_FORM TrusteeForm;
    '  TRUSTEE_TYPE TrusteeType;
    '  LPTSTR ptstrName;
    '} TRUSTEE, *PTRUSTEE;
    <StructLayout(LayoutKind.Sequential, CharSet:=CharSet.Auto)> _
    Public Structure TRUSTEE
        Dim pMultipleTrustee As IntPtr
        Dim MultipleTrusteeOperation As Integer
        Dim TrusteeForm As Integer
        Dim TrusteeType As Integer
        Dim ptstrName As String
    End Structure

    'typedef struct _EXPLICIT_ACCESS {
    '  DWORD grfAccessPermissions;
    '  ACCESS_MODE grfAccessMode;
    '  DWORD grfInheritance;
    '  TRUSTEE Trustee;
    '} EXPLICIT_ACCESS, *PEXPLICIT_ACCESS;
    <StructLayout(LayoutKind.Sequential, CharSet:=CharSet.Auto)> _
    Public Structure EXPLICIT_ACCESS
        Dim grfAccessPermissions As Integer
        Dim grfAccessMode As Integer
        Dim grfInheritance As Integer
        Dim Trustee As TRUSTEE
    End Structure

    'VOID BuildExplicitAccessWithName(
    '  PEXPLICIT_ACCESS pExplicitAccess,
    '  LPTSTR pTrusteeName,
    '  DWORD AccessPermissions,
    '  ACCESS_MODE AccessMode,
    '  DWORD Inheritance
    ');
    Declare Auto Sub BuildExplicitAccessWithName Lib "AdvAPI32.DLL" ( _
        ByRef pExplicitAccess As EXPLICIT_ACCESS, _
        ByVal pTrusteeName As String, _
        ByVal AccessPermissions As Integer, _
        ByVal AccessMode As Integer, _
        ByVal Inheritance As Integer _
    )

    'DWORD SetEntriesInAcl(
    '  ULONG cCountOfExplicitEntries,
    '  PEXPLICIT_ACCESS pListOfExplicitEntries,
    '  PACL OldAcl,
    '  PACL* NewAcl
    ');
    Declare Auto Function SetEntriesInAcl Lib "AdvAPI32.DLL" ( _
        ByVal cCountOfExplicitEntries As Integer, _
        ByRef pListOfExplicitEntries As EXPLICIT_ACCESS, _
        ByVal OldAcl As IntPtr, _
        ByRef NewAcl As IntPtr _
    ) As Integer

    'DWORD GetNamedSecurityInfo(
    '  LPTSTR pObjectName,
    '  SE_OBJECT_TYPE ObjectType,
    '  SECURITY_INFORMATION SecurityInfo,
    '  PSID* ppsidOwner,
    '  PSID* ppsidGroup,
    '  PACL* ppDacl,
    '  PACL* ppSacl,
    '  PSECURITY_DESCRIPTOR* ppSecurityDescriptor
    ');
    Declare Auto Function GetNamedSecurityInfo Lib "AdvAPI32.DLL" ( _
        ByVal pObjectName As String, _
        ByVal ObjectType As Integer, _
        ByVal SecurityInfo As Integer, _
        ByRef ppsidOwner As IntPtr, _
        ByRef ppsidGroup As IntPtr, _
        ByRef ppDacl As IntPtr, _
        ByRef ppSacl As IntPtr, _
        ByRef ppSecurityDescriptor As IntPtr _
    ) As Integer

    'DWORD SetNamedSecurityInfo(
    '  LPTSTR pObjectName,
    '  SE_OBJECT_TYPE ObjectType,
    '  SECURITY_INFORMATION SecurityInfo,
    '  PSID psidOwner,
    '  PSID psidGroup,
    '  PACL pDacl,
    '  PACL pSacl
    ');
    Declare Auto Function SetNamedSecurityInfo Lib "AdvAPI32.DLL" ( _
        ByVal pObjectName As String, _
        ByVal ObjectType As Integer, _
        ByVal SecurityInfo As Integer, _
        ByVal psidOwner As IntPtr, _
        ByVal psidGroup As IntPtr, _
        ByVal pDacl As IntPtr, _
        ByVal pSacl As IntPtr _
    ) As Integer


    Sub SetPermissions(ByRef strPath As String, ByVal strUserName As String, ByVal strPerm As String)
        Dim pNewDACL, pOldDACL As IntPtr
        Dim ea As New EXPLICIT_ACCESS
        Dim Win32Error As Win32Exception
        Dim ret, iPerm As Integer
        Dim bIsFile As Boolean

        ' Do a quick sanity check!
        bIsFile = True
        If File.Exists(strPath) = False Then
            If Directory.Exists(strPath) = False Then
                Throw New Exception("File or Directory not found")
            End If
            bIsFile = False
        End If

        ' Translate the desired permissions
        Select Case strPerm
            Case "Full Control"
                iPerm = ALL_ACCESS
            Case "Modify"
                iPerm = ALL_ACCESS And Not (WRITE_DAC Or WRITE_OWNER Or DELETE_CHILD)
            Case "Read & Execute"
                iPerm = GENERIC_READ Or GENERIC_EXECUTE
            Case "List Folder Contents"
                If bIsFile = False Then
                    Throw New Exception("Permission argument not applicable to files")
                End If
                iPerm = GENERIC_EXECUTE
            Case "Read"
                iPerm = GENERIC_READ
            Case "Write"
                iPerm = GENERIC_WRITE
            Case Else
                Throw New Exception("Invalid permission argument")
        End Select

        ' get the Security Descriptor and DACL
        ret = GetNamedSecurityInfo(strPath, SE_FILE_OBJECT, DACL_SECURITY_INFORMATION, Nothing, Nothing, pOldDACL, Nothing, Nothing)
        If ret <> 0 Then
            Win32Error = New Win32Exception(ret)
            Throw New Exception(Win32Error.Message)
        End If

        ' build an explicit access structure
        BuildExplicitAccessWithName(ea, strUserName, iPerm, SET_ACCESS, CONTAINER_INHERIT_ACE Or OBJECT_INHERIT_ACE)
        ' Oddly enough the above API doesn't work in the x64 world, so we have
        ' to do it the hard way.
        'ea.grfAccessPermissions = iPerm
        'ea.grfAccessMode = SET_ACCESS
        'ea.grfInheritance = CONTAINER_INHERIT_ACE Or OBJECT_INHERIT_ACE
        'ea.Trustee.pMultipleTrustee = IntPtr.Zero
        'ea.Trustee.MultipleTrusteeOperation = 0
        'ea.Trustee.TrusteeForm = 1
        'ea.Trustee.TrusteeType = 0
        'ea.Trustee.ptstrName = strUserName

        ' merge this Explict Access with the existing DACL
        ret = SetEntriesInAcl(1, ea, pOldDACL, pNewDACL)
        If ret <> 0 Then
            Win32Error = New Win32Exception(ret)
            Throw New Exception(Win32Error.Message)
        End If

        ' write the new Security Descriptor/DACL back
        ret = SetNamedSecurityInfo(strPath, SE_FILE_OBJECT, DACL_SECURITY_INFORMATION, Nothing, Nothing, pNewDACL, Nothing)
        If ret <> 0 Then
            Win32Error = New Win32Exception(ret)
            Throw New Exception(Win32Error.Message)
        End If

        ' clean up and go home
        Marshal.FreeHGlobal(pNewDACL)
    End Sub
End Module
